from collectionviews.iterables import ComposedIterablesBase, UnitedIterablesMixin
from collectionviews.filtered import FilteredDictsMixin
from collectionviews.sliced import DictSlicingMixin


class ComposedDictsBase(ComposedIterablesBase):

    def _generator(self):
        for source in self._sources:
            yield from source.items()

    def __iter__(self):
        return (item[0] for item in self._generator())

    def items(self):
        """Return a new view of the underlying mapping’s items ((key, value) pairs)."""
        return self._generator()

    def __contains__(self, item):
        """Return True if the underlying mappings have a key key, else False."""
        for key in self:
            if item == key:
                return True
        return False

    def __getitem__(self, key):
        """
        Return the item of the underlying mapping with key key.
        Raises a KeyError if key is not in the underlying mapping.
        """
        for _key, item in self.items():
            if _key == key:
                return item
        raise KeyError(key)

    def __len__(self):
        """Return the number of items in the underlying mapping."""
        return len(list(iter(self)))

    def copy(self):
        """Return a shallow copy of the underlying mapping."""
        pass  # todo

    def get(self, key, default=None):
        """
        Return the value for key if key is in the underlying mapping, else default.
        If default is not given, it defaults to None, so that this method never raises a KeyError.
        """
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def keys(self):
        """Return a new view of the underlying mapping’s keys."""
        return self.__iter__()

    def values(self):
        """Return a new view of the underlying mapping’s values."""
        return (item[1] for item in self._generator())


class JoinedDicts(ComposedDictsBase):

    def __getitem__(self, key):
        """
        Return a generator of the items of the underlying mappings with key key.
        NOT raises a KeyError if key is not containted in any of the underlying mappings.
        """
        return (item for _key, item in self.items() if _key == key)


class UnitedDictsMixin(object):

    def _generator(self):
        seen = set()
        for source in self._sources:
            for item in source.items():
                if item[0] not in seen:
                    yield item
                    seen.add(item[0])


class UnitedDicts(UnitedDictsMixin, ComposedDictsBase):
    pass


class FilteredJoinedDicts(FilteredDictsMixin, JoinedDicts):

    def __init__(self, filter, *sources, filters=None, conjunction=FilteredDictsMixin.ConjunctionType.AND):
        FilteredDictsMixin.__init__(self, filter, filters=filters, conjunction=conjunction)
        JoinedDicts.__init__(self, *sources)


class FilteredUnitedDicts(FilteredDictsMixin, UnitedDictsMixin, ComposedDictsBase):

    def __init__(self, filter, *sources, filters=None, conjunction=FilteredDictsMixin.ConjunctionType.AND):
        FilteredDictsMixin.__init__(self, filter, filters=filters, conjunction=conjunction)
        UnitedDicts.__init__(self, *sources)


class SlicedJoinedDicts(DictSlicingMixin, JoinedDicts):

    def __init__(self, blade, *sources, blades=None):
        DictSlicingMixin.__init__(self, blade, blades)
        JoinedDicts.__init__(self, *sources)
