from enum import Enum


class FilteredIterablesMixin(object):

    class ConjunctionType(Enum):
        AND = all
        OR = any

    def __init__(self, filter, filters=None, conjunction=all):
        super(FilteredIterablesMixin, self).__init__()
        self._filters = {filter, }
        if filters:
            self._filters.update(filters)
        self._conjunction = conjunction.value if isinstance(conjunction, FilteredIterablesMixin.ConjunctionType) \
            else conjunction

    def _generator(self):
        return filter(self._filter, super(FilteredIterablesMixin, self)._generator())

    def _filter(self, item):
        return self._conjunction(self._encapsulate_test(test)(item) for test in self._filters)

    @staticmethod
    def _encapsulate_test(test):
        def inner(*args, **kwargs):
            try:
                return test(*args, **kwargs)
            except:
                return False
        return inner


class FilteredDictsMixin(FilteredIterablesMixin):

    def _filter(self, item):
        return self._conjunction(self._encapsulate_test(test)(*item) for test in self._filters)
