from collectionviews.filtered import FilteredIterablesMixin


class ComposedIterablesBase(object):

    def __init__(self, *sources):
        super(ComposedIterablesBase, self).__init__()
        self._sources = sources

    def _generator(self):
        for item in self._sources:
            yield from item

    def __iter__(self):
        return self._generator()

    def __str__(self):
        return "<[%s]>" % ", ".join(map(str, self))

    def __repr__(self):
        return "<%s(%s)>" % (type(self).__name__, ", ".join(map(str, self._sources)))


class UnitedIterablesMixin(object):

    def _generator(self):
        seen = set()
        for item in super(UnitedIterablesMixin, self)._generator():
            if item not in seen:
                yield item
                seen.add(item)


class JoinedIterables(ComposedIterablesBase):
    pass


class UnitedIterables(UnitedIterablesMixin, ComposedIterablesBase):
    pass


class FilteredIterables(FilteredIterablesMixin, JoinedIterables):

    def __init__(self, filter, *sources, filters=None, conjunction=FilteredIterablesMixin.ConjunctionType.AND):
        FilteredIterablesMixin.__init__(self, filter, filters=filters, conjunction=conjunction)
        JoinedIterables.__init__(self, *sources)


class FilteredUnitedIterables(UnitedIterablesMixin, FilteredIterablesMixin, ComposedIterablesBase):

    def __init__(self, filter, *sources, filters=None, conjunction=FilteredIterablesMixin.ConjunctionType.AND):
        FilteredIterablesMixin.__init__(self, filter, filters=filters, conjunction=conjunction)
        ComposedIterablesBase.__init__(self, *sources)
