import copy
from .dicts import SlicedJoinedDicts, FilteredJoinedDicts


class SliceFilterWrapper(object):

    @staticmethod
    def make_filter_has_key(key, _and=None):
        if _and is None:
            _and = lambda v: True
        return lambda k, v: key in v and _and(v)

    @staticmethod
    def make_filter_has_key_eq_value(key, value):
        return SliceFilterWrapper.make_filter_has_key(key, lambda v: v[key] == value)

    SPLIT_CHAR = '|'
    ANYKEY = '*'
    FILTER_COMPARE_CHAR = '='

    def __init__(self, *sources):
        self._sources = sources

    def slice(self, item):
        return SliceFilterWrapper(SlicedJoinedDicts(item, *self._sources))

    def slice_extended(self, item):
        sources = self._sources
        item = item.split(self.SPLIT_CHAR)
        while len(item):
            key = item.pop(0)
            if key == self.ANYKEY:
                key = SlicedJoinedDicts.ANYKEY
            sources = (SlicedJoinedDicts(key, *sources), )
        return SliceFilterWrapper(*sources)

    def filter(self, filter):
        return SliceFilterWrapper(FilteredJoinedDicts(filter, *self._sources))

    def filter_extended(self, *filters):
        sources = self._sources
        for f in filters:
            pass
            attrs = f.split(self.SPLIT_CHAR)
            filter = lambda v: True
            while len(attrs) > 1:
                key = attrs.pop()
                filter = SliceFilterWrapper.make_filter_has_key(key, filter)
            key = attrs.pop()
            if self.FILTER_COMPARE_CHAR in key:
                key, value = key.split(self.FILTER_COMPARE_CHAR, 2)
                filter = SliceFilterWrapper.make_filter_has_key_eq_value(key, value)
            else:
                filter = SliceFilterWrapper.make_filter_has_key(key, filter)
            sources = (FilteredJoinedDicts(filter, *sources), )
        return SliceFilterWrapper(*sources)

    def __getitem__(self, item):
        return self.slice_extended(item)

    def __call__(self, key):
        return self.filter_extended(key)

    def items(self):
        for source in self._sources:
            yield from source.items()

    def keys(self):
        return (item[0] for item in self.items())

    def __iter__(self):
        return self.keys()

    def values(self):
        return (item[1] for item in self.items())

    def as_dict(self):
        return {item[0]: item[1] for item in self.items()}

    def update(self, E=None, **F):
        E = {} if E is None else copy.deepcopy(E)
        E.update(F)
        for key, val in self.items():
            val.update(E)
